#include "TunnelStreamClient.h"

using namespace refinitiv::ema::access;
using namespace refinitiv::ema::rdm;
using namespace std;

TunnelStreamClient::TunnelStreamClient(
	const EmaString& servicename, 
	const EmaString& username, 
	const EmaString& password, 
	const EmaString& ric) :
	_pOmmConsumer(0),
	_tunnelStreamHandle(0),
	_subStreamHandle(0),
	_serviceName(servicename),
	_rccUserName(username),
	_rccPassword(password),
	_rccPostRIC(ric),
	_BID(100),
	_ASK(102),
	_postID(1),
	_postStreamID(0)
{
	
}

TunnelStreamClient::~TunnelStreamClient()
{
	_pOmmConsumer = 0;
	_tunnelStreamHandle = 0;
	_subStreamHandle = 0;
	

}

void TunnelStreamClient::setOmmConsumer(OmmConsumer& ommConsumer)
{
	_pOmmConsumer = &ommConsumer;
}

//Handle the login refresh message sent in the tunnel stream.
//If the state of the messgae is Open/Okay, send a post message to RCC
void TunnelStreamClient::onRefreshMsg(const RefreshMsg& refreshMsg, const OmmConsumerEvent& event)
{
	cout << "TunnelStreamClient onRefreshMsg" << endl;
	cout << refreshMsg << endl;
	//Handle Login Refresh Message in the Tunnel stream
	//If the login refresh is Open/Okay, post data
	if (refreshMsg.getDomainType() == MMT_LOGIN &&
		event.getHandle() == _subStreamHandle &&
		refreshMsg.getState().getStreamState() == OmmState::OpenEnum &&
		refreshMsg.getState().getDataState() == OmmState::OkEnum)
	{
		_postStreamID = refreshMsg.getStreamId();		
		postData();

	}
}

//Handle the acknowledgement message sent in the tunnel stream.
//If the Nack flag is not set, send another post message. 
void TunnelStreamClient::onAckMsg(const AckMsg& ackMsg, const OmmConsumerEvent& event)
{
	cout << "TunnelStreamClient onAckMsg" << endl;
	cout << ackMsg << endl;
	//If the previous post message is accepted, post a new data
	if (ackMsg.hasNackCode() == false)
	{
		cout << "Posting Update....." << endl;
		postData();
	}

}

//Handle the update messages in the tunnel stream
void TunnelStreamClient::onUpdateMsg(const UpdateMsg& updateMsg, const OmmConsumerEvent& event)
{
	//Just print retrieved messages 
	cout << "TunnelStreamClient onUpdateMsg" << endl;
	cout << updateMsg << endl;
}

//Handle the status messages in the tunnel stream
//If the state of the status message in Open, send a login request message to RCC
void TunnelStreamClient::onStatusMsg(const StatusMsg& statusMsg, const OmmConsumerEvent& event)
{
	cout << "TunnelStreamClient onStatusMsg" << endl;
	cout << statusMsg << endl;
	//If the tunnel stream is opened, send a tunnel login message
	if (event.getHandle() == _tunnelStreamHandle &&
			statusMsg.hasState() &&
			statusMsg.getState().getStreamState() == OmmState::OpenEnum)
		{
			cout << "************* Login to RCC ********************" << statusMsg.getStreamId() << endl;
			loginTunnelStream(statusMsg.getStreamId());
		}
}

//Handle the generic messages in the tunnel stream
void TunnelStreamClient::onGenericMsg(const GenericMsg& genMsg, const OmmConsumerEvent& ommEvent)
{
	//Just print retrieved messages 
	cout << "TunnelStreamClient onGenericMsg" << endl;
	cout << genMsg << endl;
}


//Handle the login refresh from the AppClient
//
void TunnelStreamClient::processLoginRefresh(const refinitiv::ema::access::RefreshMsg& refreshMsg, const refinitiv::ema::access::OmmConsumerEvent& vent)
{
	cout << "TunnelStreamClient processLoginRefresh" << endl;

	//If the login state is Open/Suspect, clear the tunnnel stream handles
	 if( refreshMsg.getState().getStreamState() == OmmState::OpenEnum && refreshMsg.getState().getDataState() == OmmState::SuspectEnum)
	{
		_tunnelStreamHandle = 0;
		_subStreamHandle = 0;
	}
	

}

//Handle the login status from the AppClient
//If the login status is Open/Suspect, unset the tunnel stream
void TunnelStreamClient::processLoginStatus(const refinitiv::ema::access::StatusMsg& statusMsg, const refinitiv::ema::access::OmmConsumerEvent&)
{
	cout << "TunnelStreamClient processLoginStatus" << endl;
	//If the login state is Open/Suspect, clear the tunnnel stream handles
    if (statusMsg.hasState() && statusMsg.getState().getStreamState() == OmmState::OpenEnum && statusMsg.getState().getDataState() == OmmState::SuspectEnum)
	{
		_tunnelStreamHandle = 0;
		_subStreamHandle = 0;
	}
}

//Handle the source directory refresh from the AppClient
//If the state of the source directoy is Open/Ok, open a tunnel stream
void TunnelStreamClient::processSourceDirectoryRefresh(const refinitiv::ema::access::RefreshMsg& refreshMsg, const refinitiv::ema::access::OmmConsumerEvent&) 
{
	//If the source directory state is Open/Ok, create a tunnel stream
	cout << "TunnelStreamClient processSourceDirectoryRefresh" << endl;
	if (refreshMsg.getState().getStreamState() == OmmState::OpenEnum &&
		refreshMsg.getState().getDataState() == OmmState::OkEnum &&
		_tunnelStreamHandle == 0)
	{
		startTunnelStream();
	}

}

//Handle the source directory Update from the AppClient
void TunnelStreamClient::processSourceDirectoryUpdate(const refinitiv::ema::access::UpdateMsg& updateMsg, const refinitiv::ema::access::OmmConsumerEvent&) {
	cout << "TunnelStreamClient processSourceDirectoryUpdate" << endl;

	//create a tunnel stream if it is not created
	if (_tunnelStreamHandle == 0) {
		startTunnelStream();
	}
}

void TunnelStreamClient::processSourceDirectoryStatus(const refinitiv::ema::access::StatusMsg&, const refinitiv::ema::access::OmmConsumerEvent&) {
	//Print Log	
	cout << "TunnelStreamClient processSourceDirectoryStatus" << endl;
}

//Create a tunnel stream
void TunnelStreamClient::startTunnelStream() {
	cout << "Start Tunnel Stream" << endl;
	CosAuthentication cosAuthentication;
	cosAuthentication.type(CosAuthentication::NotRequiredEnum);

	CosDataIntegrity cosDataIntegrity;
	cosDataIntegrity.type(CosDataIntegrity::ReliableEnum);

	CosFlowControl cosFlowControl;
	cosFlowControl.type(CosFlowControl::BidirectionalEnum).recvWindowSize(12288).sendWindowSize(1200);

	CosGuarantee cosGuarantee;
	cosGuarantee.type(CosGuarantee::NoneEnum);

	ClassOfService cos;
	cos.authentication(cosAuthentication).dataIntegrity(cosDataIntegrity).flowControl(cosFlowControl).guarantee(cosGuarantee);

	TunnelStreamRequest tsr;
	tsr.classOfService(cos).domainType(MMT_SYSTEM).name("Tunnel").serviceName(_serviceName);

	_tunnelStreamHandle = _pOmmConsumer->registerClient(tsr, *this);
}


//Login to the Tunnel Stream
void TunnelStreamClient::loginTunnelStream(refinitiv::ema::access::Int32 streamId) {
	
	ElementList elmList;
	elmList.addAscii("Password", _rccPassword).complete();

	ReqMsg reqMsg;

	reqMsg.domainType(MMT_LOGIN)
		.name(_rccUserName)
		//.privateStream(true)
		.attrib(elmList)
		.streamId(streamId);
	cout << endl << "Submit GenericMsg from onStatus" << endl;

	_subStreamHandle = _pOmmConsumer->registerClient(reqMsg, *this, (void*)1, _tunnelStreamHandle);
	
}
//Send a post message to RCC
void TunnelStreamClient::postData() {
	cout << "Send First PostMsg()....." << endl;
	_pOmmConsumer->submit(PostMsg()
		.streamId(_postStreamID)
		.postId(_postID).domainType(MMT_MARKET_PRICE)
		.solicitAck(true).complete()
		.payload(UpdateMsg().streamId(_postStreamID)
			.name(_rccPostRIC)
			.payload(FieldList()
				.addReal(22, _BID, OmmReal::ExponentNeg2Enum)
				.addReal(25, _ASK, OmmReal::ExponentNeg2Enum)
				.complete()))
		.complete(),
		_subStreamHandle);

	_postID++;
	_BID++;
	_ASK++;
}